#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
test_recipe
----------------------------------

Tests for `inicheck.entries` module.
"""

import unittest
from inicheck.entries import RecipeSection, TriggerEntry
from collections import OrderedDict

class TestEntries(unittest.TestCase):
    def test_trigger_entry(self):
        """
        Tests to see if we correctly gather a trigger for a recipe
        """
        entries = [ 'has_section = test',
                    'has_item = test',
                    'has_value = [test test test]',
                    ['has_section = test',
                    'has_item = test2']]

        # section trigger
        t = TriggerEntry(entries[0])
        assert(t.conditions[0] == ['test', 'any', 'any'])

        # section item
        t = TriggerEntry(entries[1])
        assert(t.conditions[0] == ['any', 'test', 'any'])

        t = TriggerEntry(entries[2])
        assert(t.conditions[0] == ['test', 'test', 'test'])

        # Confirm we can handle multiple conditions
        t = TriggerEntry(entries[3])
        assert(t.conditions[0] == ['test', 'any', 'any'])
        assert(t.conditions[1] == ['any', 'test2', 'any'])


    def test_config_entry(self):
        """
        Tests to see if we correctly gather a trigger for a recipe
        """
        entries = [['has_value = [topo type ipw]']]

        t = TriggerEntry(entries[0])
        assert(t.conditions[0] == ['topo', 'type', 'ipw'])

if __name__ == '__main__':
    import sys
    sys.exit(unittest.main())



# !/usr/bin/env python

import argparse
from . output import print_config_report, generate_config, print_recipe_summary, print_details, print_non_defaults
from . tools import get_user_config, check_config
import os
import sys
from . config import MasterConfig

def main():

    parser = argparse.ArgumentParser(description="Examine and auto populate"
                                                 " ini files with a master file"
                                                 " comparison.")

    parser.add_argument('--config_file','-f', dest='config_file', type=str,
                        help='Path to a config file that needs checking')

    parser.add_argument('--master', '-mf', metavar='MF', type=str, nargs='+',
                        help='Path to a config file that used to check against')

    parser.add_argument('--modules', '-m', metavar='M', type=str, nargs='+',
                    help="Modules name with an attribute __CoreConfig__ that"
                         " is a path to a master config file for checking"
                         " against")

    parser.add_argument('--write','-w', dest='write', action='store_true',
                        help="Determines whether to write out the file with all"
                             " the defaults")

    parser.add_argument('--recipes','-r', dest='recipes', action='store_true',
                        help="Prints out the recipe summary")

    parser.add_argument('--non-defaults','-nd', dest='defaults',
                        action='store_true',
                        help="Prints out a summary of the non-defaults")

    parser.add_argument('--details', '-d', type=str, nargs='+', help="Provide"
                        " section item and value for details regarding them")


    args = parser.parse_args()

    # Module not provided, or master config
    if args.modules == None and args.master == None:
        print("ERROR: Please provide either a module or a path to a master"
             " config, or ask for details on config entries")
        sys.exit()


    # Normal operation
    else:
        # Requesting details for a section and item
        if args.details != None:
            if len(args.details) == 1:
                print("Providing details for section {0}..."
                "".format(args.details[0]))

            elif len(args.details) == 2:
                print("Providing details for section {0} and item {1}..."
                      "".format(args.details[0], args.details[1]))

            else:
                print("Details option can at most recieve section and item ")
                sys.exit()

            mcfg = MasterConfig(path=args.master, modules=args.modules)
            print_details(args.details, mcfg.cfg)

        # Requesting a check on a config file
        else:
            f = os.path.abspath(args.config_file)
            ucfg = get_user_config(f, master_files=args.master,
                                      modules=args.modules,
                                      checking_later=True)

            warnings, errors = check_config(ucfg)

            print_config_report(warnings, errors)

            # Print out the recipes summary
            if args.recipes:
                print_recipe_summary(ucfg.recipes)

            # Print out the summary of non-defaults values
            if args.defaults:
                print_non_defaults(ucfg)

            # Output the config file as inicheck interprets it.
            if args.write:
                out_f = './{0}_full.ini'.format(
                                              os.path.basename(f).split('.')[0])

                print("Writing complete config file with all recipes and "
                      " necessary defaults...")
                print('{0}'.format(out_f))

                generate_config(ucfg, out_f, cli=True)

if __name__ == '__main__':
    main()



from collections import OrderedDict
from . iniparse import parse_entry
from . import __trigger_keywords__

class RecipeSection:
    """
    The RecipeSection enables the master config file to parses a section
    that is actually a recipe. Operates using keywords provided n the name
    of the section. It collects triggers and adjustments to the config file
    and assigns them to this class. Trigger represent the conditional
    statements to apply the adjustments listed  below. If an item in the
    section doesn't contain any trigger keyword then it is assumed to be
    an adjustment to be made.

    """

    def __init__(self, recipe_section_dict, name=None):
        """
            Args:
                recipe_section_dict: a dictionary containing the triggers
                                     and config file adjustments.
        """
        self.name = name

        # Conditions to be met
        self.triggers = OrderedDict()

        # Config file to apply if conditions are met
        self.adj_config = OrderedDict()

        for item, entry in recipe_section_dict.items():

            # Check item for action keywords
            for word in __trigger_keywords__:
                if word in item:
                    self.triggers[item] = TriggerEntry(entry)
                    break

            # Check for assigned values if any trigger
            if item not in self.triggers.keys():
                item_dict = parse_entry(entry, item = item)

                self.adj_config[item] = item_dict

class TriggerEntry:
    """
    RecipeEntry designed to aid in parsing master config file entries under
    a recipe.
    This is meant to parse:

    Example::

        ------------------------------------------------------------
        item_trigger:
                        has_section_name = <value>,
                        has_value = [<section name> <item name>, <value>],
                        has_item_name = <value>

        -------------------------------------------------------------

    Config entry expects to recieve the above in the following format:

    .. code-block:: python

        {item_trigger:
                    ["has_section = <value>",
                     "has_item = [<section> <item>"],
                     "has_value = [<section> <item> <value>]"
                     ]
        }

    Recipe entry then will parse the strings looking for space separated lists,
    values denoted with = and will only accept keyword

        * has_section
        * has_item
        * has_value

    """

    def __init__(self, parseable_line, name=None):

        self.conditions = []
        self.valid_names = ['has_section', 'has_item', 'has_value']
        heirarcy = ['section', 'item', 'value']

        parsed_dict = parse_entry(parseable_line,
                                  valid_names=self.valid_names)
        # There can be multiple conditions returned
        for name,value in parsed_dict.items():
            result = ['any', 'any', 'any']
            if type(value) == list:
                # Easy assignment to result using [section  item value syntax]
                for i,v in enumerate(value):
                    result[i] = v

            # If single item provided
            else:
                for i,keyword in enumerate(heirarcy):
                    if keyword in name:
                        result[i] = value

            # If result is all any, then clear it
            if len([True for i in result if i == 'any']) != len(result):
                self.conditions.append(result)


class ConfigEntry:
    """
    ConfigEntry designed to aid in parsing master config file entries.
    This is meant to parse:

    Example::

        ------------------------------------------------------------
        item:
                        type = <value>,
                        options = [<value> <value>],
                        description = text describing entry
        -------------------------------------------------------------

    Config entry expects to recieve the above in the following format:

    .. code-block:: python

        {item:
            ["type = <value>",
             "options = [<value> <value>"],
             "description=text describing entry"]
        }

    Config entry then will parse the strings looking for space separated
    lists, values denoted with =, and will only receive:

        * type
        * default
        * options
        * description

    """

    def __init__(self, name=None, value=None, default=None,
                 entry_type='string', options=[], parseable_line=None):

        self.name = name
        self.value = value
        self.default = default
        self.options = options
        self.description = ''
        self.listed = False
        self.type = entry_type
        self.valid_names = ['default', 'type', 'options', 'description']

        if parseable_line != None:
            parsed_dict = parse_entry(parseable_line, item = name,
                                        valid_names=self.valid_names)
            for name,value in parsed_dict.items():
                setattr(self,name,value)

        # Options should always be a list
        if type(self.options) != list:
            self.options = [self.options]

        # types should always be lower case
        self.type = self.type.lower()

        # Handle the list types
        if 'list' in self.type:
            self.listed = True
            self.type = self.type.replace('list','')
            self.type = self.type.strip()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
test_config
----------------------------------

Tests for `inicheck.config` module.
"""

import unittest
import os
from inicheck.config import *
from inicheck.tools import cast_all_variables

def compare_config(generated_config, truth_config, master=False, mcfg_attr='default'):
    """
    Compares config objects for quick tests:

    Args:
        generated_config: the config file the inicheck system produced automatically
        truth_config: a dictionary made up for checking against
        master: boolean for identifying if were checking master config
        mcfg_attr: attribute we want to check when using a master config
    Returns:
        result: boolean describing if cfg are the same.
    """

    # First confirm all keys are available
    for s in truth_config.keys():
        if s  not in generated_config:
            print("\nERROR: test data contains section: {} that was not"
                  " contained by the generated config.".format(s))
            return False

        for i,v in truth_config[s].items():

            # Then confirm the item is in the section in the generated config
            if i not in generated_config[s].keys():
                print("\nERROR: test data contains item: {} in section: {} that was not"
                      " contained by the generated config.".format(i,s))
                return False

            # Check the value
            else:
                # if it is a master config object then use it differently.
                if master:
                    gv = getattr(generated_config[s][i], mcfg_attr)
                    emsg = "\nERROR: config.{} = {} not {}.".format(
                                                    mcfg_attr,
                                                    gv,
                                                    v)

                # Normal config
                else:
                    gv = generated_config[s][i]
                    emsg ="\nERROR: config[{}][{}] = {} not {}.".format(s,i,gv,
                                                                              v)
                # Confirm values are the same
                if gv != v:
                        print(emsg)
                        return False

    return True


class TestUserConfig(unittest.TestCase):
    def setUp(self):
        """
        Stage truth data here, grab the config to pass around
        """

        fname = os.path.abspath(os.path.dirname(__file__)+'/test_configs/base_cfg.ini')

        mcfg = MasterConfig(modules = 'inicheck')
        self.ucfg = UserConfig(fname, mcfg=mcfg)

    def test_base_config(self):
        """
        Simply opens and looks at the base config BEFORE all the recipes and
        stuff has been applied
        """
        truth = {'topo':
                        {},
                 'air_temp':
                        {}
                }

        assert compare_config(self.ucfg.cfg,truth)

    def test_apply_recipes(self):
        """
        Tests that the correct variables are in place after recipes
        """

        truth = {'topo':
                        {'filename':None,
                        'type':'netcdf'},
                 'air_temp':
                        {'distribution': 'idw',
                         'detrend':'true',
                         'slope':'1'}
                }

        self.ucfg.apply_recipes()
        assert compare_config(self.ucfg.cfg,truth)

    def test_apply_casting(self):
        """
        Tests that the correct variables are in place after recipes
        """

        truth = {'topo':
                        {'filename':None,
                        'type':'netcdf'},
                 'air_temp':
                        {'distribution': 'idw',
                         'detrend':True,
                         'slope':1.0}
                }

        self.ucfg.apply_recipes()
        self.ucfg = cast_all_variables(self.ucfg, self.ucfg.mcfg)

        assert compare_config(self.ucfg.cfg,truth)
        

class TestMasterConfig(unittest.TestCase):
    def setUp(self):
        """
        Stage our truthing data here
        """
        self.truth_defaults = {'topo':
                                            {'type': 'netcdf',
                                             'dem': None,
                                             'filename': None},
                                'air_temp':
                                            {'distribution': 'idw',
                                             'detrend': 'true',
                                             'dk_nthreads': '1'}}

    def test_from_module(self):
        """
        Builds a master config from the module, check it.
        """
        # Build a master config file using multiple files
        mcfg = MasterConfig(modules = 'inicheck')
        assert compare_config(mcfg.cfg,self.truth_defaults, master=True,
                                                            mcfg_attr='default')

    def test_from_paths(self):
        """
        Builds a master config from the files, check it.
        """
        # Build a master config file using multiple files
        mcfg = MasterConfig(path = ['./tests/test_configs/CoreConfig.ini',
                                         './tests/test_configs/recipes.ini'])

        assert compare_config(mcfg.cfg,self.truth_defaults, master=True,
                                                            mcfg_attr='default')


if __name__ == '__main__':
    import sys
    sys.exit(unittest.main())



import os
from .utilities import  remove_chars, remove_comment
from collections import OrderedDict


def read_config(fname):
    """
    Opens and reads in the config file in its most raw form.
    Creates a dictionary of dictionaries that contain the string result

    Args:
        fname: Real path to the config file to be opened
    Returns:
        config: dict of dicts containing the info in a config file
    """
    with open(fname, encoding = 'utf-8') as f:
        lines = f.readlines()
        f.close()

    sections = parse_sections(lines)
    sec_and_items = parse_items(sections)
    config = parse_values(sec_and_items)

    return config


def parse_entry(info, item = None, valid_names=None):
    properties = OrderedDict()
    if type(info) != list:
        info = [info]
    last_three = []

    for s in info:
        if '=' in s:
            a = s.split('=')

        else:
            raise ValueError('\n\nMaster Config file missing an equals sign in'
                            ' entry or missing a comma right before the item '
                            '"{0}" in the entry:\n"{1}"'.format(item,info))

        name = (a[0].lower()).strip()
        # Check for constraints on potential names of entries
        if valid_names != None and name not in valid_names:
            raise ValueError("Invalid option set in the master config File"
                            " for entry -----> {0}".format(name))

        value = a[1].strip()

        result = []
        value = value.replace('\n'," ")
        value = value.replace('\t',"")

        # Is there a list of values provided?
        if '[' in value:
            if ']' not in value:
                raise ValueError('Missing bracket or commas used in a list'
                                 ' instead of spaces in config file under'
                                 ' {0} in the entry:\n"{1}"'.format(name, info))
            else:
                value = (''.join(c for c in value if c not in '[]'))
                value = value.split(' ')

        properties[name] = value
    return properties


def parse_sections(lines):
    """
    Returns a dictionary containing all the sections as keys with a single
    string of the contents after the section

    Args:

        lines: a list of string lines containing all the raw info of a cfg file

    Returns:
        sections: dictionary containing keys that are the section names and
                  values that are strings of the contents between sections

    """

    result = OrderedDict()
    section = None
    i = 0

    for i in range(len(lines)):
        line = remove_chars(lines[i],'\t')
        # Comment checking
        line = remove_comment(line)

        # Watch out for extraneous chars at the beginning an end
        line = line.strip()

        # Check for empty line first
        if line and line not in os.linesep:
            # Look for section
            if line.startswith('['):
                # Look for open brackets
                if ']' in line:

                    section = (remove_chars(line,'[]')).lower()
                    result[section] = []

            else:
                # This protects from funky syntax in a config file and alerts the user
                if section == None:
                    raise Exception("Non-section like syntax before any "
                                    "sections were identified at line {0} in "
                                    "config file. Please use bracketed sections"
                                    " or use # or ; to write comments."
                                    "".format(i))
                else:
                    result[section].append(line)

    return result


def parse_items(parsed_sections_dict, mcfg=None):
    """
    Takes the output from parse_sections and parses the items in each
    section

    Args:
        parsed_sections_dict: dictionary containing keys as the sections
                              and values as a continuous string of the
    Returns:
        result: dictionary of dictionaries containing sections,items, and
                their values
    """
    result = OrderedDict()

    for k,v in parsed_sections_dict.items():
        item = None
        result[k] = OrderedDict()
        for i,val in enumerate(v):
            val = val.lstrip()
            # Look for item notation

            if ':' in val:
                # Only split on the first colon to avoid collisions with datetime
                parse_loc = val.index(':')
                parseable = [val[0:parse_loc],val[parse_loc + 1:]]
                item = parseable[0].lower()

                result[k][item] = ''

                # Check for a value right after the item name
                potential_value = (parseable[-1].replace('\n', ' ')).lstrip()

                # Avoid stashing properties in line with the item
                if '=' not in potential_value:
                    result[k][item] = potential_value

                # Property value provided in line with item
                else:
                    result[k][item] += " "+potential_value

            # User added line returns likely for readability
            else:
                result[k][item] += " "+val

        # Perform a final cleanup
        if item != None:
            if ',' in result[k][item]:
                result[k][item] = ", ".join([e.strip() for e in result[k][item].split(',')])

            result[k][item] = result[k][item].strip()

    return result


def parse_values(parsed_items):
        """
        Takes the output from parse_items and parses any values or
        properties provided in each item placing the strings into a list.
        If no properties are defined then it will clean up values provided
        and put them in a list of len one.

        Args:
            parsed_sections_dict: dict of dicts containing joined strings
                                  found under each item
        Returns:
            result: dictionary of dictionaries containing sections, items,
                    and the values provided as a list
        """
        result = OrderedDict()
        for section in parsed_items.keys():
            result[section] = OrderedDict()
            for item,val in parsed_items[section].items():
                value = val.strip()
                if ',' in val:
                    result[section][item] = \
                    [v.strip() for v in value.split(',')]

                else:
                    result[section][item] = [value]

        return result



import os
import sys
import inspect
from . config import UserConfig, MasterConfig
from . utilities import mk_lst
import inicheck.checkers

def get_checkers(module='inicheck.checkers', keywords=[]):
    """
    Args:
        module: The module to search for the classes

    Returns a dictionary of the classes available for checking config entries
    """
    keywords.append('check')

    funcs = inspect.getmembers(sys.modules[module], inspect.isclass)
    func_dict = {}
    for name, fn in funcs:
        checker_found = False
        k = name.lower()
        # Remove any keywords
        for w in keywords:
            z = w.lower()
            if z in k:
                k = k.replace(z, '')
                checker_found = True
                break

        if checker_found:
            func_dict[k] = fn

    return func_dict

def check_config(config_obj):
            """
            Looks at the users provided config file and checks it to a master
            config file looking at correctness and missing info.

            Args:
                config_obj - UserConfig object produced by
                             :class:`~inicheck.config.UserConfig`
            Returns:
                tuple:
                - **warnings** - Returns a list of string messages that are
                                 consider non-critical issues with config file.
                - **errors** - Returns a list of string messages that are
                               consider critical issues with the config file.
            """

            msg = "{: <20} {: <30} {: <60}"
            errors = []
            warnings = []

            master = config_obj.mcfg.cfg
            cfg = config_obj.cfg

            standard_funcs = get_checkers()

            # Add any checker modules if provided
            if config_obj.mcfg.checker_modules:
                for c in config_obj.mcfg.checker_modules:
                    funcs = standard_funcs.update(
                            get_checkers(module=c))

            # Compare user config file to our master config
            for section, configured in cfg.items():

                if section not in master.keys():
                    err = "Not a valid section."
                    errors.append(msg.format(section, " ", err))
                else:
                    # In the section check the values and options
                    for item, value in configured.items():
                        litem = item.lower()

                        # Is the item known as a configurable item?
                        if litem not in master[section].keys():
                            wrn = "Not a registered option."
                            warnings.append(msg.format(section, item, wrn))
                        else:
                            # Did the user provide a list value or single value
                            val_lst = mk_lst(value)

                            #Check to see if we want to print the item error
                            # location in the list
                            if len(val_lst) > 1:
                                print_lst = True
                            else:
                                print_lst = False

                            for ii,v in enumerate(val_lst):
                                if v != None:

                                    # If we care about the errors position,
                                    # print it
                                    if print_lst:
                                        print_item = "{0}[{1}]".format(item,ii)
                                    else:
                                        print_item = item

                                    # 1. Check for contraints by options lists
                                    if master[section][item].options:
                                        # If it is not in the list, invalid
                                        if str(v) not in master[section][item].options:
                                            full_msg = msg.format(section,
                                                              print_item,
                                                              "Not a valid option")
                                            errors.append(full_msg)

                                    # 2. Check the type constraint.
                                    options_type = master[section][item].type

                                    for name, fn in standard_funcs.items():
                                        # Check for type checkers
                                        if options_type == name.lower():
                                            b = fn(value=v, config=config_obj,
                                           is_list=master[section][item].listed)
                                            issue = b.check()

                                            if issue != None:
                                                full_msg = msg.format(section,
                                                                      print_item,
                                                                      issue)
                                                if b.msg_level == 'error':
                                                    errors.append(full_msg)
                                                elif b.msg_level == 'warning':
                                                    warnings.append(full_msg)
                                                break
                                        else:
                                            issue = None
            return warnings, errors


def cast_all_variables(config_obj, mcfg_obj, checking_later = False):
    """
    Cast all values into the appropiate type using checkers, other_types
    and the master config.

    Args:
        config_obj: The object of the user config from
        mcfg_obj: The object used for manage the master config from
                  class MasterConfig
        other_types: User provided list to add any custom types
        checking_later: Enables whether a failure to cast an item will raise an exception

    Returns:
        ucfg: The users config dictionary containing the correct value types

    """

    ucfg = config_obj.cfg
    mcfg = mcfg_obj.cfg
    all_checks = get_checkers()

    # Add any checker modules if provided
    if config_obj.mcfg.checker_modules:
        for c in config_obj.mcfg.checker_modules:
            new_checks = get_checkers(module=c)
            all_checks.update(new_checks)

    # Cast all variables
    for s in ucfg.keys():
        if s in mcfg.keys():
            for i in ucfg[s].keys():
                values = []

                # Ensure it is something we can check
                if i in mcfg[s].keys():
                    type_value = (mcfg[s][i].type).lower()

                    if type_value not in all_checks.keys():
                        raise ValueError("\n\nSection {0} at item {1} attempted"
                                        " to use undefined type name '{2}'"
                                        " which has no checker associated."
                                        "\nAvailable checkers "
                                        "are:\n\n{3}"
                                        "".format(s,i,type_value,all_checks.keys()))

                    for z, v in enumerate(mk_lst(ucfg[s][i])):
                        option_found = False

                        # when a checker match is found break so we check it once
                        for name, fn in all_checks.items():
                            # Checker name and type match
                            if  type_value == name.lower():
                                b = fn(value=v, config=config_obj)

                                # Be wary of the None
                                if v in [None, 'none', 'None']:
                                    values.append(None)

                                else:
                                    # cfg will be checked later all at once
                                    if checking_later:
                                        try:
                                            values.append(b.cast())

                                        except:
                                            values.append(v)

                                    else:
                                        values.append(b.cast())

                                option_found = True
                                break

                        if not option_found:
                            raise ValueError("Unknown type_value prescribed."
                                             " ----> {0}".format(type_value))

                    # Developers can specify which items are lists in core cfg
                    if not mcfg[s][i].listed or (len(values)==1 and values[0]==None):
                        ucfg[s][i] = mk_lst(values, unlst=True)

                # Not recognized items, keep them anyways
                else:
                    values.append(ucfg[s][i])

    config_obj.cfg = ucfg
    return config_obj


def get_user_config(config_file, master_files=None, modules=None,
                    mcfg=None, checking_later=False):
    """
    Returns the users config as the object UserConfig.

    Args:
        config_file: real path to existing config file
        master_file: real path to a Core Config file
        modules: a module or list of modules with a string attribute __CoreConfig__ which is the
                path to a CoreConfig
        mcfg: the master config object after it has been read in.
        checking_later: Passes over excpetions when catsing to the right types to be formally checked later

    Returns:
        ucfg: Users config as an object
    """

    if modules == None and master_files == None and mcfg == None:
        raise IOError("ERROR: Please provide either a module or a path to a"
                      " master config, or a master config object")
        sys.exit()

    if os.path.isfile(config_file):

        if master_files != None or modules != None:
            if master_files != None:
                master_files = mk_lst(master_files)

            mcfg = MasterConfig(path=master_files, modules=modules)

        ucfg = UserConfig(config_file, mcfg=mcfg)

        ucfg.apply_recipes()
        ucfg = cast_all_variables(ucfg, mcfg, checking_later=checking_later)

    else:
        raise IOError("Config file path {0} doesn't exist."
                      "".format(config_file))

    return ucfg

def config_documentation(out_f, paths=None, modules=None,
                        section_link_dict={}):
    """
    Auto documents the core config file. Outputs to a file which is can then be
    used for documentation. Specifically formulated for sphinx

    Args:
        out_f: string path to output location for the auto documentation
        paths: paths to master config files to use for creating docs
        modules: modules with attributes __core_config__ for creating docs
        section_link_dict- dictionary containing special documentation for a section in the config file reference

    """

    if paths == None and modules == None:
        raise ValueError("inicheck function config_documentation args paths or"
                        " module must be specified!")


    master = MasterConfig(path=paths, modules=modules)
    mcfg = master.cfg

    # Beginning
    config_doc = "\nFor configuration file syntax information please visit"
    config_doc += " http://inicheck.readthedocs.io/en/latest/\n\n"

    for section in mcfg.keys():
        # Section header
        config_doc += "\n"
        config_doc += "{0}\n".format(section)
        config_doc += "-" * len(section) + '\n'

        # Allow for custom info in a description about the section, useful
        # for linking modules for auto documenting
        if section in section_link_dict.keys():
            intro = section_link_dict[section]
        else:
            intro = ''

        config_doc += intro
        config_doc += "\n"

        # Auto document config file according to master config contents
        for item,v in sorted(mcfg[section].items()):
            # Check for attributes that are lists
            for att in ['default', 'options']:
                z = getattr(v,att)
                if type(z) == list:
                    combo = ' '
                    doc_s = combo.join([str(s) for s in z])
                    setattr(v,att,doc_s)

            # Bold item with definition
            config_doc += "| **{0}**\n".format(item)

            # Add the item description
            config_doc += "| \t{0}\n".format(v.description)

            # Default
            config_doc += "| \t\t*Default: {0}*\n".format(v.default)

            # Add expected type
            config_doc += "| \t\t*Type: {0}*\n".format(v.type)

            # Print options should they be available
            if v.options:
                config_doc += "| \t\t*Options:*\n *{0}*\n".format(v.options)

            config_doc += "| \n"

            config_doc += "\n"

    path = os.path.abspath(out_f)
    if not os.path.isdir(os.path.dirname(path)):
        raise IOError("Config Documentation output file does not exist."
                      "\n{0}".format(out_f))

    print("Writing auto documentation for config file to:\n{0}".format(path))

    with open(path,'w+') as f:
        f.writelines(config_doc)
    f.close()



# -*- coding: utf-8 -*-
import os
__author__ = """Micah Johnson"""
__email__ = 'micah.johnson150@gmail.com'
__version__ = '0.3.2'

# Keywords for conditional statements in a master config file
__trigger_keywords__ = ['trigger']

# Key words expected in the section name of a recipe
__recipe_keywords__ = ['recipe']

# Add in core configs for test purposes
__core_config__ = os.path.abspath(os.path.dirname(__file__)+'/../tests/test_configs/CoreConfig.ini')
__recipes__ = os.path.abspath(os.path.dirname(__file__)+'/../tests/test_configs/recipes.ini')



'''
Functions for checking valus in a config file and producing errors and warnings
'''
import os
from pandas import to_datetime
from .utilities import mk_lst

class GenericCheck(object):
    def __init__(self, **kwargs):

        if 'value' not in kwargs.keys():
            raise ValueError("Must provided at least keyword value to "
                             "Checkers.")
        if 'config' not in kwargs.keys():
            raise ValueError("Must provided at least keyword config to"
                             " Checkers.")

        self.message = None
        self.msg_level = 'warning'
        self.value = kwargs["value"]

        self.config = kwargs["config"]

        if not self.msg_level.lower() in ['warning', 'error']:
            raise ValueError("msg_level = {0} not allowed."
                             "".format(self.msg_level))

    def is_valid(self):
        """
        Abstract function for defining how a value is checked for validity

        Args:
            value: Value that is going to be check for validity
        Returns:
            **(boolean, msg)**: True or False whether value valid and
                                correspond error message if it is not
        """
        pass

    def check(self):
        """
        Function that is ran by the checking function of the user config to
        return useful message to instruct user

        Args:
            value: value to be check for validity
        Returns:
            msg: None is the entry is valid, else returns self.message
        """

        valid, issue = self.is_valid()
        if valid:
            msg = None
        else:
            msg = issue
        return msg


class CheckType(GenericCheck):
    """
    Base class for type checking whether a value of the right type.
    """

    def __init__(self, **kwargs):
        super(CheckType, self).__init__(**kwargs)
        self.type = 'str'
        # Function used for casting to types
        self.type_func = str

        # Allow users to specify an option to be a list
        if 'is_list' in kwargs.keys():
            self.is_list=kwargs['is_list']
        else:
            self.is_list = False

    def is_it_a_lst(self):
        """
        Check to see if an option should be a list of not and convert it if
        necessary.
        If it is a list, check for single value lists and convert it, return false.
        If it is a list and not single item, return True
        if its not a list return false

        """
        if type(self.value) == list:
            # Its list but its a single item
            if len(self.value) == 1:

                self.value = mk_lst(self.value, unlist=True)
                return False

            # Its a list and not a single item
            else:
                return True

        # not a list, we good
        else:
            return False

    def is_valid(self):
        """
        Checks for type validity

        Returns:
        tuple:
            valid: Boolean whether the value was acceptable
            msg: string to print if value is not valid.
        """
        msg = None
        self.msg_level = 'error'

        # Provide a list check
        currently_a_list = self.is_it_a_lst()

        if currently_a_list and not self.is_list:
            msg = "Expected single value received list"
            valid = False
        else:

            try:

                self.value = self.cast()
                valid = True

            except:
                msg = "Expecting {0} received {1}".format(self.type,
                                                    type(self.value).__name__)
                valid = False

        return valid,msg

    def cast(self):
        """
        Attempts to return the casted value
        """
        return self.type_func(self.value)


class CheckDatetime(CheckType):
    """
    Check values that are declared as type datetime.
    """

    def __init__(self, **kwargs):

        super(CheckDatetime, self).__init__(**kwargs)
        self.type_func = to_datetime
        self.type = 'datetime'


class CheckFloat(CheckType):
    """
    Float checking whether a value of the right type.
    """

    def __init__(self, **kwargs):

        super(CheckFloat, self).__init__(**kwargs)
        self.type_func = float
        self.type = 'float'


class CheckInt(CheckType):
    """
    integer checking whether a value of the right type.
    """

    def __init__(self, **kwargs):

        super(CheckInt, self).__init__(**kwargs)
        self.type_func = self.cast_float_int
        self.type = 'int'


    def cast_float_int(self,value):
        """
        When expecting an integer, it is convenient to automatically convert
        floats to integers (e.g. 6.0 --> 6) but its pertinent to catch when the
        input has a non-zero decimal and warn user (e.g. avoid 6.5 --> 6)

        Args:
            value: The value to be casted to integer
        Returns:
            self.value: the value converted
        """

        self.value = float(value)

        if self.value.is_integer():
            self.value = int(self.value)
        else:
            raise ValueError("Expecting integer and received float with non-zero decimal")

        return self.value


class CheckBool(CheckType):
    """
    Boolean checking whether a value of the right type.
    """

    def __init__(self, **kwargs):

        super(CheckBool, self).__init__(**kwargs)
        self.type_func = bool
        self.type = 'bool'
        self.affirmatives = ['y', 'yes', 'true']
        self.negatives = ['n', 'no', 'false']

    def cast(self):

        v = str(self.value).lower()
        if v in self.affirmatives:
            self.value = True

        elif v in self.negatives:
            self.value = False
        else:
            raise ValueError("Value {0} not coercable to boolean."
                             "".format(self.value))

        return self.value


class CheckString(CheckType):
    """
    Float checking whether a value of the right type.
    """

    def __init__(self, **kwargs):

        super(CheckString, self).__init__(**kwargs)
        self.type_func = str
        self.type = 'string'


class CheckPath(CheckType):
    """
    Checks whether a Path exists.
    """

    def __init__(self, **kwargs):

        super(CheckPath, self).__init__(**kwargs)
        self.root_loc = self.config.filename

        self.dir_path = False

        # Path should alsways be absolute or relative to the config file path
        if self.value != None and self.root_loc != None:
            if not os.path.isabs(self.value):
                p = os.path.expanduser(os.path.dirname(self.root_loc))
                self.value = os.path.abspath(os.path.join(p, self.value))

    def is_valid(self):
        """
        Checks for existing filename
        """

        if self.dir_path:
            exists = os.path.isdir(self.value)
        else:
            exists = os.path.isfile(self.value)

        return exists,self.message

    def cast(self):
        return self.value


class CheckDirectory(CheckPath):
    """
    Checks whether a directory exists.
    """

    def __init__(self, **kwargs):
        super(CheckDirectory, self).__init__(**kwargs)
        self.dir_path = True
        self.message = "Directory does not exist."


class CheckFilename(CheckPath):
    """
    Checks whether a directory exists.
    """

    def __init__(self, **kwargs):
        super(CheckFilename, self).__init__(**kwargs)
        self.message = "File does not exist."


class CheckCriticalFilename(CheckFilename):
    """
    Checks whether a critical file exists.
    """

    def __init__(self, **kwargs):
        super(CheckCriticalFilename, self).__init__(**kwargs)
        self.msg_level = 'error'


class CheckCriticalDirectory(CheckDirectory):
    """
    Checks whether a critical directory exists.
    """

    def __init__(self, **kwargs):
        super(CheckCriticalDirectory, self).__init__(**kwargs)
        self.msg_level = 'error'



import sys
import os

def remove_comment(string_entry):
    """
    Takes a single line and removes and .ini type comments.
    Also remove any spaces at the begining or end of a commented line
    """
    # Comment checking
    if '#' in string_entry:
        result = string_entry.split('#')[0]
    elif ';' in string_entry:
        result = string_entry.split(';')[0]
    else:
        result = string_entry

    return result


def mk_lst(values, unlst=False):
    """
    while iterating through several type of lists and items it is convenient to
    assume that we recieve a list, use this function to accomplish this. It also
    convenient to be able to return the original type after were done with it.
    """
    # Not a list, were not looking to unlist it. make it a list
    if type(values) != list and not unlst:
            values = [values]
    else:
        # We want to unlist a list we get
        if unlst:
            # single item provided so we don't want it a list
            if len(values) == 1:
                values = values[0]

    return values


def remove_chars(orig_str, char_str, replace_str=None):
    """
    Removes all character in orig_str that are in char_str

    Args:
        orig_str: Original string needing cleaning
        char_str: String of characters to be removed
        replace_str: replace values with a character if requested

    Returns:
        string: orig_str with out any characters in char_str
    """
    if replace_str == None:
        clean_str = [c for c in orig_str if c not in char_str]
    else:
        clean_str = [c if c not in char_str else replace_str for c in orig_str]

    return "".join(clean_str)


def get_relative_to_cfg(path, user_cfg_path):
    """
    Converts a path so that all paths are relative to the config file

    Args:
        path: relative str path to be converted
        user_cfg_path: path to the users config file
    Returns:
        path: absolute path of the input considering it was relative to the cfg
    """
    if not os.path.isabs(path):
        path = os.path.abspath(os.path.join(os.path.dirname(user_cfg_path),
                                                            path))
    return path


def pcfg(cfg):
    """
    prints out the config file to the prompt with a nice stagger Look for
    readability

    Args:
        cfg: dict of dict in a heirarcy of  {section, {items, values}}
    """

    for sec in cfg.keys():
        print(repr(sec))
        try:
            for item in cfg[sec].keys():
                print('\t' + item)
                values = cfg[sec][item]

                if type(cfg[sec][item]) == list:
                    out = ", ".join(cfg[sec][item])
                else:
                    out = cfg[sec][item]
                print('\t\t' + repr(out))
        except:
            print('\t recipe')


def pmcfg(cfg):
    """
    prints out the core config file to the prompt with a nice stagger Look for
    readability

    Args:
        cfg: dict of dict in a heirarcy of  {section, {items, values}}
    """

    for sec in cfg.keys():
        print(repr(sec))
        for item in cfg[sec].keys():
            print('\t' + item)
            obj = cfg[sec][item]
            for att in ['type', 'options', 'default', 'description']:
                value = getattr(obj, att)
                print('\t\t' + att)

                if type(value) == list:
                    out = ", ".join(value)
                else:
                    out = value
                print('\t\t\t' + repr(out))



from .iniparse import read_config
from .utilities import mk_lst, get_relative_to_cfg
from .entries import ConfigEntry, RecipeSection
from . import __recipe_keywords__
from collections import OrderedDict
import os
import copy
import importlib
import copy

DEBUG = False
FULL_DEBUG = False

class UserConfig():
    """
    Class meant for managing the the users config, here we operate on the
    config repeatedly making it available through the attribute self.cfg
    """

    def __init__(self, filename, mcfg=None):
        """
        Args:
            filename: String to path containing config in .ini format
            mcfg: Object of the master config
        """
        self.filename = filename
        self.recipes = []

        # Hang on to the original
        self.raw_cfg = read_config(filename)

        # Version inicheck will mess with
        self.cfg = copy.deepcopy(self.raw_cfg)

        self.sections, self.items, self.values = \
        self.get_unique_entries(self.cfg)

        if mcfg != None:
            self.mcfg = mcfg

    def apply_recipes(self):
        """
        Look through the users config file and section by section add in
        missing parameters to add defaults

        Returns:
            user_cfg: User config dictionary with defaults added.
        """
        # Add this in case the user has added anything to the config obj on the fly
        self.cfg = copy.deepcopy(self.raw_cfg)

        # Start fresh with recipes to avoid over populating the recipes list
        self.recipes = []

        for r in self.mcfg.recipes:

            for trigger, recipe_entry in r.triggers.items():
                conditions_met = 0
                triggered = False

                # All conditions must be met if to be applied
                for condition in recipe_entry.conditions:
                    conditions_triggered = []
                    for section in self.cfg.keys():

                        # Watch out for empty sections
                        if len(self.cfg[section].keys()) == 0:
                            items = [None]
                        else:
                            items = self.cfg[section].keys()

                        for item in items:

                            # Watch for empties
                            if item == None:
                                vals = [None]
                            else:
                                vals = mk_lst(self.cfg[section][item])
                            for v in vals:

                                if (condition[0] == 'any' or
                                    condition[0] == section):
                                   if FULL_DEBUG:
                                       print("Section Gate {0} == {1}"
                                             "".format(condition[0],
                                             section))

                                   if (condition[1] == 'any' or
                                       condition[1] == item):
                                      if FULL_DEBUG:
                                          print("\t\tItem Gate {0} == {1}"
                                                "".format(condition[1],
                                                          item))

                                      if (condition[2] == 'any' or
                                          condition[2] == v):
                                          if FULL_DEBUG:
                                             print("\t\t\t\tValue Gate {0}"
                                                   " == {1}"
                                                   "".format(condition[2],
                                                             v))

                                          # No conditions == [any any any]
                                          conditions_triggered.append(
                                                        (section, item, v))
                                          triggered = True

                    # Determine if the condition was met.
                    if triggered:
                        conditions_met += 1
                        triggered = False

                if (conditions_met == len(recipe_entry.conditions) and
                    len(recipe_entry.conditions) != 0):
                    conditions_met = 0
                    self.recipes.append(r)

                    # Check recipes for sections not in the master
                    invalid_r_sections = [s for s in r.adj_config.keys()
                                          if (s not in self.mcfg.cfg.keys()
                                             and s != 'any')]

                    if len(invalid_r_sections) > 0:
                        raise ValueError("The recipe {} attempts to modify"
                        " section(s) {} not recognized by Master Config."
                        "".format(r.name,",".join(invalid_r_sections)))

                    if DEBUG:
                        print("\nDEBUG: Trigger: {0} {1} was met!"
                        "".format(trigger, condition))

                    # Iterate through the conditions found and apply
                    for situation in conditions_triggered:
                        # Insert the recipe into the users config
                        self.cfg = self.interpret_recipes(r.adj_config,
                                                          situation)

                else:
                    if DEBUG:
                        print("\nDEBUG: Trigger: {0} not met. gates = {1}"
                              " and gates_passed = {2}"
                              "".format(trigger, condition,
                                        conditions_met))
                        print('\n\n')


    def interpret_recipes(self, partial_cfg, situation):
        """
        User inserts a partial config by using each situation that
        triggered a recipe. A triggering situation consists of a tuple of
        (section, item, value) that represent the specific settings that trigger
        the recipe.
        """
        result = copy.deepcopy(self.cfg)

        for section in partial_cfg.keys():
            for item in partial_cfg[section].keys():
                remove = False
                value = partial_cfg[section][item]
                values = mk_lst(value)

                for value in values:

                    # Defaults Keyword
                    if item == 'apply_defaults':
                        if str(value).lower() == 'true':
                            result = self.add_defaults(result, sections=section)

                    # Keyword removal
                    elif item == 'remove_section':
                        if value.lower() == 'true':

                            if section in result.keys():
                                if DEBUG:
                                    print("removed section: {0}"
                                          "".format(section))
                                del result[section]

                    # Normal operation
                    else:

                        # Handle the any keyword for sections
                        if section == 'any':
                            s = situation[0]

                        else:
                            s = section
                        # Handle the any keyword for items
                        if item == 'any':
                            i = situation[1]

                        # Observe the description shift
                        elif item in ['remove_item','default_item']:
                            i = value

                        else:
                            i = item

                        if value == 'any':
                            v = situation[2]

                        elif (value == 'default' or item =='default_item'):
                            if i in self.mcfg.cfg[s].keys():
                                v = self.mcfg.cfg[s][i].default
                            else:
                                raise Exception('{0} is not a valid item for '
                                'the master config section {1}, check your '
                                'recipes for a situation triggering on {2}, '
                                '{3}, {4}'.format(i,s,situation[0],situation[1],
                                                                  situation[2]))

                        else:
                            v = value

                        # Delete items
                        if item == 'remove_item':
                            if s in result.keys():
                                if i in result[s].keys():
                                    if DEBUG:
                                        print("Removed: {0} {1}".format(s, i))
                                    del result[s][i]

                        elif s in result.keys():
                            # Check for empty dictionaries
                            if not bool(result[s]):
                                if DEBUG:
                                    print("Adding section {0}".format(s))
                                result[s] = OrderedDict()
                            # Dictionary exists
                            else:

                                # Handle a item not provided by automatically adding it
                                if i not in result[s].keys():
                                    if DEBUG:
                                        print("Adding {0} {1} {2}"
                                              "".format(s, i, v))
                                    result[s][i] = v

                                # If the item was provided we don't want to overide the user with defaults
                                elif value != 'default':
                                    if DEBUG:
                                        print("Changing {0} {1} {2}"
                                              "".format(s, i, v))
                                    result[s][i] = v

        return result


    def get_unique_entries(self, cfg):
        """
        Appends all the values in the user config to respectives lists of
        section names, item names, and values. Afterwards any copy is
        removed so all is left is a unique list of names and values
        """

        unique_sections = []
        unique_items = []
        unique_values = []

        for section in cfg.keys():
            for item, value in cfg[section].items():
                if type(value) != list:
                    vals = [value]
                else:
                    vals = value

                for v in vals:
                    unique_sections.append(section)
                    unique_items.append(item)
                    unique_values.append(v)

        return set(unique_sections), set(unique_items), set(unique_values)


    def add_defaults(self, cfg, sections=None,items=None):
        """
        Look through the users config file and section by section add in
        missing parameters to add defaults

        Args:
            sections: Single section name or a list of sections to apply
                      (optional) otherwise uses all sections in users
                      config
        Returns:
            user_cfg: User config dictionary with defaults added.

        """
        master = self.mcfg.cfg
        result = copy.deepcopy(cfg)
        #Either go through specified sections or all sections provided by the user.
        if sections == None:
            sections = result.keys()
        else:
            #Accounts for single items not entered as a list
            sections = mk_lst(sections)

        for section in sections:
            configured = result[section]
            for k, v in master[section].items():
                if v.name not in configured.keys():
                    result[section][k] = v.default
        return result

    def update_config_paths(self, user_cfg_path=None):
            """
            Sets all paths so that they are always relative to the config
            file or absolute.
            """
            if user_cfg_path == None:
                user_cfg_path = self.filename

            mcfg = self.mcfg.cfg
            cfg = self.cfg

            #Cycle thru users config
            for section in cfg.keys():
                for item in cfg[section].keys():
                    d = cfg[section][item]
                    #Does master have this and is it not none
                    if item in mcfg[section].keys() and d != None:
                        m = mcfg[section][item]
                        #Any paths
                        if m.type == 'filename' or  m.type == 'directory':
                            cfg[section][item] = \
                            get_relative_to_cfg(cfg[section][item],
                                                self.filename)
            return cfg


class MasterConfig():
    def __init__(self, path=None, modules=None, checkers=None, titles=None,
                 header=None):

        if type(path) != list:
            path = [path]
        self.paths = []
        self.recipes = []
        self.titles = {}
        self.header = header

        self.checker_modules = []

        # Paths were manually provided
        if path != None and modules == None:
            for p in path:
                self.paths.append(os.path.abspath(p))

        # If a module was passed
        if modules != None and self.paths == []:
            if type(modules) != list:
                modules = [modules]

            for m in modules:
                i = importlib.import_module(m)
                self.paths.append(os.path.abspath(os.path.join(i.__file__,
                                                         i.__core_config__)))

                # Search for possible recipes provided in the module
                if hasattr(i, '__recipes__'):
                    self.paths.append(i.__recipes__)

                # Search for possible section titles provided in the module
                if hasattr(i, '__config_titles__'):
                    self.titles.update(getattr(i, '__config_titles__'))

                # Search for config headers provided in the module
                if hasattr(i, '__config_header__'):
                    self.header = getattr(i, '__config_header__')

                # Search for custom checkers
                if hasattr(i, '__config_checkers__'):
                    self.checker_modules.append(m+'.' +
                                              getattr(i, '__config_checkers__'))

        if checkers != None:
            if type(checkers) != list:
                checkers = [checkers]
            for c in checkers:
                self.checker_modules.append(c)

        if titles != None:
            self.titles.update(titles)

        if header != None:
            self.header = header

        if len(self.paths) == 0 and modules == None:
            raise ValueError("No file was either provided or found when"
                             " initiating a master config file.")

        self.cfg = self.add_files(self.paths)

    def add_files(self, paths):
        """
        Designed to  add to the master config file if the user has split
        up files to reduce the amount of info in a master file. e.g.
        recipes are stored in another file.

        Args:
            paths: list of real path to another cfg.ini
        Returns:
            config: Original config with appended information found in the
                    extra cfg
        """

        result = OrderedDict()

        for f in paths:
            if f != None:
                extra_cfg = self._read(f)
                result.update(extra_cfg)

        return result

    def merge(self,mcfg):
        """
        Merges the a master config object into the current master config object
        in place
        """

        self.recipes+=mcfg.recipes
        self.checker_modules += mcfg.checker_modules
        self.titles.update(mcfg.titles)
        self.cfg.update(mcfg.cfg)

    def _read(self, master_config_file):
        """
        Reads in the core config file which has special syntax for
        specifying options

        Args:
            master_config_file: String path to the master config file.

        Returns:
            config: Dictionary of dictionaries representing the defaults
                    and available options. Based on the Core Config file.
        """

        cfg = OrderedDict()

        # Read in will automatically get the configurable key added
        raw_config = read_config(master_config_file)

        for section in raw_config.keys():
            sec = OrderedDict()
            for word in __recipe_keywords__:
                if word in section:
                    self.recipes.append(RecipeSection(raw_config[section], name=section))
                    break

                else:
                    for item in raw_config[section].keys():

                        sec[item] = ConfigEntry(name=item,
                                  parseable_line=raw_config[section][item])

                    cfg[section] = sec

        return cfg



from datetime import date
import os
import sys
from . utilities import mk_lst

def generate_config(config_obj, fname, cli=False):
    """
    Generates a list of strings using the config data then its written to an
    .ini file

    Args:
        config_obj: config object containing data to be outputted
        fname: String path to the output location for the new config file
        cli: Boolean value that adds the line "file generated using
            inicheck.cli", Default = False


    """
    header_len = 80
    pg_sep = '#' * header_len

    # Header surround each commented titles in the ini file
    section_header = pg_sep + '\n' + ('# {0}\n') + pg_sep

    # Construct the section strings
    config_str = ""
    config_str += pg_sep

    # File header with specific package option
    if config_obj.mcfg.header != None:
        header = config_obj.mcfg.header.split('\n')
        for line in header:
            config_str += ('\n# ' + line)
    else:
        config_str += "\n# Configuration File "

    # Add in the date generated
    config_str += "\n#\n# Date generated: {0}".format(date.today())

    # Generated with inicheck
    if cli:
        config_str += "\n#\n# Generated using: inicheck <filename> -w"

    config_str += "\n#\n# For more inicheck help see:" + \
                  "\n# http://inicheck.readthedocs.io/en/latest/\n"

    config = config_obj.cfg
    mcfg = config_obj.mcfg.cfg

    # Check to see if section titles were provided
    has_section_titles = hasattr(config_obj.mcfg,'titles')

    # Generate the string for the file, creating them in order.
    for section in mcfg.keys():
        if section in config.keys():
            config_str += '\n' * 2

            # Add a section header
            s_hdr = pg_sep
            if has_section_titles:
                if section in config_obj.mcfg.titles.keys():
                    # Add the header
                    s_hdr = section_header.format(config_obj.mcfg.titles[section])

            else:
                config_str += s_hdr

            config_str += s_hdr
            config_str += '\n'
            config_str += '\n[{0}]\n'.format(section)

            # Add section items and values
            for k in config[section].keys():
                v = config[section][k]
                if type(v) == list:
                    astr = ", ".join(str(c).strip() for c in v)
                else:
                    astr = str(v)
                config_str += "{0:<30} {1:<10}\n".format((k+':'), astr)

    # Write out the string generated
    with open(os.path.abspath(fname), 'w') as f:
        f.writelines(config_str)
        f.close()


def print_config_report(warnings, errors, logger=None):
    """
    Pass in the list of string messages generated by check_config file.
    print out in a pretty format the issues

    Args:
        warnings: List of non-critical messages returned from
                   :func:`~utilities.check_config'.
        errors: List of critical messages returned from
                 :func:`~utilities.check_config'.
        logger: pass in the logger function being used. If no logger is
                 provided, print is used. Default = None

    """

    msg = "{: <20} {: <25} {: <60}"

    # Check to see if user wants the logger or stdout
    if logger != None:
        out = logger.info
    else:
        out = print_out


    msg_len = 90
    out(" ")
    out(" ")
    out("Configuration File Status Report:")
    header = "="*msg_len
    out(header)
    any_warnings = False
    any_errors = False

    # Output warnings
    if len(warnings) > 0:
        any_warnings = True
        out("WARNINGS:")
        out(" ")
        out(msg.format(" Section", "Item", "Message"))
        out("-"*msg_len)
        for w in warnings:
            out(w)
        out(" ")
        out(" ")

    # Output errors
    if len(errors) > 0:
        any_errors = True
        out("ERRORS:")
        out(" ")
        out(msg.format("Section", "Item", "Message"))
        out("-"*msg_len)
        for e in errors:
            out(e)
        out(" ")
        out(" ")

    if not any_errors and not any_warnings:
        out("No errors or warnings were reported with the config file.\n")


def print_out(out_str):
    """
    wrapper for print so we can use either a logger or a stdout
    """
    print(out_str)


def print_recipe_summary(lst_recipes):
    """
    Prints out the recipes found and how they are interpretted

    Args:
        lst_recipes: list of the recipe entry objects

    """
    # len of recipe separators
    msg_len = 80
    header = "=" * msg_len
    recipe_hdr = "-" * msg_len
    r_msg = "\n{0: <20}\n"+recipe_hdr
    cfg_msg = "\t\t{0: <20} {1: <20} {2: <20}"

    msg = "\t\t{0: <20} {1: <25}"

    print('\n\n')
    print("Below are the recipes applied to the config file:")
    print("Recipes Summary:")
    print(header)

    for r in lst_recipes:
        print(r_msg.format(r.name))
        print("\tConditionals:")
        for n, t in r.triggers.items():
            for i, c in enumerate(t.conditions):
                if type(c) == list:
                    c = ", ".join(c)
                if i == 0:
                    print(msg.format(n, c))
                else:
                    print(msg.format("", c))

        print_cfg_for_recipe(r.adj_config, cfg_msg, hdr="\n\tEdits:")
        #print('\n')
    print('\n')


def print_cfg_for_recipe(cfg, fmt, hdr=None):
    if hdr != None:
        print(hdr)

    for section in cfg.keys():
        for item, value in cfg[section].items():
            if type(value) != list:
                v = [value]
            else:
                v = value
                
            for qq in v:
                print(fmt.format(section, item, qq))


def print_details(details, mcfg):
    """
    Prints out the details for a list of provided options designed for use
    with the CLI. Details about a section, or an item can be requested by
    passing in a list of in the section,item order. If a section is only passed
    then we the details provided are for the entire section

    Args:
        details: a list in [section item value] requesting details.
        mcfg: master config object to gather the details from


    """

    msg = "{: <15} {: <15} {: <15} {: <25} {: <60}"
    hdr  = '\n' + msg.format('Section', 'Item', 'Default', 'Options',
                           'Description')
    print(hdr)
    print('='*len(hdr))
    nopts = len(details)
    # At least a section was provided
    if nopts >= 1:
        if details[0] in mcfg.keys():
            # A section and item was provided
            if nopts == 2:
                if details[1] in mcfg[details[0]].keys():
                    print(msg.format(details[0], details[1],
                                    str(mcfg[details[0]][details[1]].default),
                                    str(mcfg[details[0]][details[1]].options),
                                    str(mcfg[details[0]][details[1]].description)))
                else:
                    print("Item {0} in not a registered item."
                          "".format(details[1]))
                    sys.exit()

            # Print the whole section
            else:
                for k, v in mcfg[details[0]].items():

                    print(msg.format(details[0],
                                     k,
                                     str(v.default),
                                     str(v.options),
                                     str(v.description)))

        # Section does not exist
        else:
            print("Section {0} in not a valid section.".format(details[0]))
            sys.exit()

    else:
        print("Please provide at least a section for information")
        sys.exit()


def print_non_defaults(ucfg):
    """
    Prints out the options used that were not default option values.

    Args:
        ucfg: config object containing options that are not default
    """

    mcfg = ucfg.mcfg.cfg
    cfg = ucfg.cfg

    msg = "{: <20} {: <20} {: <40} {: <40}"
    hdr = '\n'+msg.format("Section","Item","Value","Default")

    print("\n\nConfiguration File Non-Defaults Report:")
    print("The following are all the items that had non-defaults values specified.")
    print("="*len(hdr))
    print(hdr)
    print('-'*len(hdr))

    # Cycle through option/items checking defaults, print em if they don't match
    for s in mcfg.keys():
        if s in cfg.keys():
            for i in mcfg[s].keys():
                if i in cfg[s].keys():
                    default_lst = mk_lst(mcfg[s][i].default)
                    str_default_lst =  [str(kk).lower() for kk in default_lst]
                    user_lst = mk_lst(cfg[s][i])
                    str_lst = [str(kk).lower() for kk in user_lst]

                    for vi, v in enumerate(str_default_lst):
                        # Single entries
                        if v != 'none':
                            if len(str_lst) == 1:
                                if str_lst[vi] != v:
                                    print(msg.format(s, i,str(user_lst[vi]),
                                                          str(default_lst[vi])))

                            # Handle awkward list comparisons
                            elif v not in str_lst:
                                print(msg.format(s, i,"all values",
                                                       str(default_lst[vi])))

    print("")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('docs/history.rst') as history_file:
    history = history_file.read()

with open('./requirements.txt') as req_file:
    requirements = req_file.read()


setup_requirements = [
    'pytest-runner',
    # TODO(micahjohnson150): put setup requirements (distutils extensions, etc.) here
]

test_requirements = [
    'pytest',
    # TODO: put package test requirements here
]

setup(
    name='inicheck',
    version='0.3.2',
    description="inicheck is an high level configuration file checker enabling developers tight control over their users configuration files",
    long_description=readme + '\n\n' + history,
    author="Micah Johnson",
    author_email='micah.johnson150@gmail.com',
    url='https://github.com/micahjohnson150/inicheck',
    packages=find_packages(include=['inicheck']),
    entry_points={
        'console_scripts': [
            'inicheck=inicheck.cli:main'
        ]
    },
    include_package_data=True,
    install_requires=requirements,
    license="GNU General Public License v3",
    zip_safe=False,
    keywords='inicheck',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    setup_requires=setup_requirements,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
test_iniparse
----------------------------------

Tests for `inicheck.iniparse` module.
"""

import unittest
from inicheck.iniparse import parse_sections, parse_items, parse_entry, parse_values


class TestIniparse(unittest.TestCase):

    def test_parse_sections(self):
        """
        Tests our base function used to parse sections before we do any
        processing

        Ensures:

        * we can handle comments around sections.
        * caps issues

        """

        info = ["# Test our line comments",
                "[unit]",
                "a:10; in line comment",
                "[test CASe] # another inline comment test",
                "b:5"]
        sections = parse_sections(info)

        # Confirm we see a normal section
        assert(sections['unit'][0] == "a:10")

        # Confirm we see a section with space caps issues
        assert(sections['test case'][0] == "b:5")

        # Confirm we catch dumb errors before we find a section name
        self.assertRaises(Exception, parse_sections,['a#','#','[test]'])


    def test_parse_items(self):
        """
        Tests our base function used to parse items after we read sections but
        before processing values.

        * tests for wrapped lists
        * tests to remove bracketed lists with comments
        * tests for properties being added in master files
        """

        info = {"unit":["a:10"],
                "test case":["a:10,","15,20"],
                'recipe':['a: default=10,',
                          'options=[10 15 20]']} # Handle wrapped lines with tabs

        items = parse_items(info)

        # Check for a normal single entry parse
        #print(items['test case']['a'])
        assert(items['unit']['a'] == '10')

        # Check for correct interpretation of a wrapped list
        assert(items['test case']['a'] == "10, 15, 20")

        # Check for a correct interpretation of item properties for master files
        assert(items['recipe']['a'] == "default=10, options=[10 15 20]")

    def test_parse_values(parsed_items):
        """
        test parse values

        Ensures:
        * we can clean up values with list entries
        * we can clean up properties we recieve including non-comma sep lists
        """
        info = {'unit':{'a': 'test1,\ttest2, test3'},
                'recipe':{'a':'\tdefault=10,\toptions=[10 15 20]'}
                }

        values = parse_values(info)

        # Check we parse a comma list correctly
        assert(values['unit']['a'][1] == 'test2')

        # Check we parse a properties list with no commas correctly
        assert(values['recipe']['a'][1]=='options=[10 15 20]')


if __name__ == '__main__':
    import sys
    sys.exit(unittest.main())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
test_utilities
----------------------------------

Tests for `inicheck.utilities` module.
"""

import unittest
from inicheck.utilities import *


class TesUtilities(unittest.TestCase):

    def test_remove_comments(self):
        """
        Test our remove comments code
        """
        values = {"test":"test#comment",
                  "test":"test;comment"}
        for k,v in values.items():
            out = remove_comment(v)

            assert k==out

    def test_mk_lst(self):
        """
        Test all the ways we use mk_lst

        Case A: Convert non-list to list and keep lists as lists
        Case B: Convert lists to single value and keep singles as single
        Case C: Don't undo lists that are acutal lists

        """

        # Case A
        for v in ['test',['test']]:
            out = mk_lst(v,unlst=False)
            assert type(out) == list

        # Case B
        for v in ['test',['test']]:
            out = mk_lst(v,unlst=True)
            assert type(out) != list

        # Case C
        for v in [['test','test2']]:
            out = mk_lst(v,unlst=True)
            assert type(out) == list

    def test_remove_chars(self):
        """
        Test if we can remove the problematic chars
        """
        out = remove_chars("\t\n my_test\t","\t\n ")
        assert '\n' not in out
        assert '\t' not in out
        assert ' ' not in out


if __name__ == '__main__':
    import sys
    sys.exit(unittest.main())
